@@ -7,7 +7,7 @@
 File created: September 4th 2020
 
 Modified By: hsky77
-Last Updated: January 3rd 2021 09:26:34 am
+Last Updated: January 24th 2021 15:55:44 pm
 '''
 
 from uuid import UUID
@@ -253,6 +253,13 @@ class ClauseCompiler():
                     val = compiled._bind_processors[key](raw_val)
                 else:
                     val = raw_val
+                if hasattr(query, 'table') and key in query.table.primary_key.columns:
+                    if not hasattr(query, 'select_pk_clauses'):
+                        query.select_pk_clauses = query.table.primary_key.columns[key] == val
+                    else:
+                        query.select_pk_clauses = and_(
+                            query.select_pk_clauses, query.table.primary_key.columns[key] == val)
+
                 args.append(val)
 
             execution_context.result_column_struct = (
@@ -264,7 +271,8 @@ class ClauseCompiler():
 
     def set_default_value(self, clause: ClauseElement, context: ExecutionContext):
         for column in clause.table.columns:
-            if not column.name in clause.parameters and column.default:
+            if (column.name in clause.parameters and clause.parameters[column.name] is None and column.default) or (
+                    not column.name in clause.parameters and column.default):
                 clause.parameters[column.name] = column.default.arg(context) if callable(
                     column.default.arg) else column.default.arg
 
@@ -363,16 +371,15 @@ class AioSqliteCursorProxy(AsyncCursorProxy):
     async def execute_return_row(self, clause: ClauseElement) -> Optional[RowProxy]:
         query, args, _ = self._compiler.compile(clause)
         await self.cursor.execute(query, args)
-        rowid = self.cursor.rowcount if self.cursor.lastrowid == 0 else self.cursor.lastrowid
-
-        # fetch inserted/updated row
-        await self.cursor.execute('SELECT * from {} where rowid = {}'.format(clause.table.name, rowid))
-        row = await self.cursor.fetchone()
-        _, _, context = self._compiler.compile(clause.table.select())
-        if row is None:
-            return None
-        metadata = ResultMetaData(context, self.cursor.description)
-        return RowProxy(metadata, row, metadata._processors, metadata._keymap)
+        if clause.table._autoincrement_column is not None:
+            rowid = self.cursor.rowcount if self.cursor.lastrowid == 0 else self.cursor.lastrowid
+            select_clause = clause.table.select().where(
+                clause.table._autoincrement_column == rowid)
+            row = await self.fetch_one(select_clause)
+        else:
+            select_clause = clause.table.select().where(clause.select_pk_clauses)
+            row = await self.fetch_one(select_clause)
+        return row
 
     async def execute_many(self, queries: List[ClauseElement]) -> None:
         for query in queries:
@@ -443,16 +450,15 @@ class AioMysqlCursorProxy(AsyncCursorProxy):
     async def execute_return_row(self, clause: ClauseElement) -> Optional[RowProxy]:
         query, args, _ = self._compiler.compile(clause)
         await self.cursor.execute(query, args)
-        rowid = self.cursor.rowcount if self.cursor.lastrowid == 0 else self.cursor.lastrowid
-
-        # fetch inserted/updated row
-        await self.cursor.execute('SELECT * from {} where rowid = {}'.format(clause.table.name, rowid))
-        row = await self.cursor.fetchone()
-        _, _, context = self._compiler.compile(clause.table.select())
-        if row is None:
-            return None
-        metadata = ResultMetaData(context, self.cursor.description)
-        return RowProxy(metadata, row, metadata._processors, metadata._keymap)
+        if clause.table._autoincrement_column is not None:
+            rowid = self.cursor.rowcount if self.cursor.lastrowid == 0 else self.cursor.lastrowid
+            select_clause = clause.table.select().where(
+                clause.table._autoincrement_column == rowid)
+            row = await self.fetch_one(select_clause)
+        else:
+            select_clause = clause.table.select().where(clause.select_pk_clauses)
+            row = await self.fetch_one(select_clause)
+        return row
 
     async def execute_many(self, queries: List[ClauseElement]) -> None:
         for query in queries:
@@ -475,6 +481,11 @@ class AioMysqlCursorProxy(AsyncCursorProxy):
 
 
 class AsyncConnectionProxy():
+    def __init__(self, async_rdb: "AsyncSQLAlchemyRDB"):
+        self._async_rdb = async_rdb
+        self.reconnect = False
+        self.acquire_count = 0
+
     async def __aenter__(self):
         await self.acquire()
         return self
@@ -490,19 +501,22 @@ class AsyncConnectionProxy():
         raise NotImplementedError()
 
     async def acquire(self):
-        raise NotImplementedError()
+        self.acquire_count += 1
 
     async def get_cursor_proxy(self) -> AsyncCursorProxy:
         raise NotImplementedError()
 
     def release(self):
-        raise NotImplementedError()
+        self.acquire_count -= 1
+
+    def set_reconnect(self):
+        self.reconnect = True
 
 
 class AioMysqlConnectionProxy(AsyncConnectionProxy):
-    def __init__(self,
+    def __init__(self, async_rdb: "AsyncSQLAlchemyRDB",
                  host: str, port: int, user: str, password: str, db_name: str):
-        super().__init__()
+        super().__init__(async_rdb)
         self.host = host
         self.port = port
         self.user = user
@@ -528,7 +542,12 @@ class AioMysqlConnectionProxy(AsyncConnectionProxy):
         self.release()
 
     async def acquire(self, ping: bool = True):
+        await super().acquire()
         await self._lock.acquire()
+        if self.reconnect and hasattr(self, '_conn'):
+            self._conn.close()
+            self._is_connected = False
+
         if not self._is_connected:
             self._is_connected = True
             self._conn = await aiomysql.connect(
@@ -538,11 +557,12 @@ class AioMysqlConnectionProxy(AsyncConnectionProxy):
 
     def release(self):
         self._lock.release()
+        super().release()
 
 
 class AioSqliteConnectionProxy(AsyncConnectionProxy):
-    def __init__(self, file_name: str):
-        super().__init__()
+    def __init__(self, async_rdb: "AsyncSQLAlchemyRDB",  file_name: str):
+        super().__init__(async_rdb)
         self._lock = Lock()
         self._conn = aiosqlite.connect(file_name)
         self._is_connected = False
@@ -559,13 +579,19 @@ class AioSqliteConnectionProxy(AsyncConnectionProxy):
         return AioSqliteCursorProxy(self)
 
     async def acquire(self):
+        await super().acquire()
         await self._lock.acquire()
+        if self.reconnect and hasattr(self, '_conn'):
+            self._conn.close()
+            self._is_connected = False
+
         if not self._is_connected:
             self._is_connected = True
             await self._conn
 
     def release(self):
         self._lock.release()
+        super().release()
 
     async def dispose(self):
         await self.acquire()
@@ -605,14 +631,32 @@ class AioMySQLDatabase(AsyncSQLAlchemyRDB):
         self.password = kwargs['password']
         self.db_name = kwargs['db_name']
 
-        self.connection_proxy = AioMysqlConnectionProxy(
-            self.host, self.port, self.user, self.password, self.db_name)
+        self.max_connections = kwargs.get('connections', 5)
+        self.pool = []
 
     def get_connection_proxy(self) -> AioMysqlConnectionProxy:
-        return self.connection_proxy
+        conn_proxy = None
+        for conn in self.pool:
+            if conn.acquire_count == 0:
+                return conn
+
+        if len(self.pool) < self.max_connections:
+            conn_proxy = AioMysqlConnectionProxy(
+                self, self.host, self.port, self.user, self.password, self.db_name)
+            self.pool.append(conn_proxy)
+            return conn_proxy
+        else:
+            for conn in self.pool:
+                if conn_proxy is None:
+                    conn_proxy = self.pool[0]
+                elif conn_proxy.acquire_count > conn.acquire_count:
+                    conn_proxy = conn
+
+            return conn_proxy
 
     async def dispose(self):
-        await self.connection_proxy.dispose()
+        for conn in self.pool:
+            await conn.dispose()
 
 
 class AioSqliteDatabase(AsyncSQLAlchemyRDB):
@@ -624,7 +668,7 @@ class AioSqliteDatabase(AsyncSQLAlchemyRDB):
         super().__init__(DB_MODULE_NAME.SQLITE_FILE, declared_entity_base, connect_args,
                          engine_args, **kwargs)
         self.connection_proxy = AioSqliteConnectionProxy(
-            kwargs['file_name'])
+            self, kwargs['file_name'])
 
     def get_connection_proxy(self) -> AioSqliteConnectionProxy:
         return self.connection_proxy
@@ -669,6 +713,8 @@ class AsyncEntityUW():
             return str_to_datetime(v).date()
         elif issubclass(t, enum.IntEnum):
             return t(int(v))
+        elif v is None:
+            return None
         else:
             return t(v)
 
